// Palette animation

#define _GNU_SOURCE

#include <err.h>
#include <stdlib.h>
#include <libxml/parser.h>
#include <ocell/core.h>
#include <ocell/graphic.h>
#include <ocell/super_cast.h>

typedef struct cycle {
	struct cycle *next;
	unsigned position;
	int speed;
	int start;
	int end;
	unsigned next_step;
} cycle;

typedef struct {
	unsigned cycles;
	cycle *cycle;
} cycle_init;

#define KEY_HANDLER_CAST(p, t) SUPER_CAST(p, t, key_handler, process, p)

typedef struct key_handler key_handler;

#define PALANIM_CAST(p, t) SUPER_CAST(p, t, palanim, process, p)

typedef struct palanim palanim;
typedef struct {
	_Bool do_cycle;
	cycle *cycle;
	cycle **cycle_order;
	int ncolors;
	SDL_Color *colors;
	SDL_Color *original_colors;
	unsigned ticks;
} palanim_private;

#define VIEWER_CAST(p, t) SUPER_CAST(p, t, viewer, process, p)

typedef struct viewer viewer;

enum {
	FPS = 50
};

static const char *palanim_path;
static size_t private_offset;

static key_handler *key_handler_create(process *father);
static void key_handler_run(process *p);
static size_t key_handler_sizeof(void);
static palanim *palanim_create(process *father, const char *path);
static palanim_private *palanim_get_private(palanim *pa);
static void palanim_on_kill(process *p);
static void palanim_run(process *p);
static size_t palanim_sizeof(void);
static void process_args(int argc, char **argv);
static void viewer_run(process *p);
static size_t viewer_sizeof(void);

int main(int argc, char **argv) {
	process_args(argc, argv);
	private_offset = process_sizeof();
	viewer *v = alloc(viewer_sizeof());
	process *p = VIEWER_CAST(v, process);
	program_init(p);
	p->run = viewer_run;
	run_program();
	return EXIT_SUCCESS;
}

static void add_cycle(xmlNodePtr node, cycle_init *ci) {
	if (xmlNodeIsText(node))
		node = node->next;
	if (!xmlStrEqual(node->name, (const xmlChar *) "speed"))
		return;
	xmlNodePtr content_node = node->children;
	int speed = strtol((char *) content_node->content, NULL, 10);
	if (speed < 1)
		return;
	node = node->next;

	if (xmlNodeIsText(node))
		node = node->next;
	if (!xmlStrEqual(node->name, (const xmlChar *) "start"))
		return;
	content_node = node->children;
	int start = strtol((char *) content_node->content, NULL, 10);
	if (start < 0)
		return;
	node = node->next;

	if (xmlNodeIsText(node))
		node = node->next;
	if (!xmlStrEqual(node->name, (const xmlChar *) "end"))
		return;
	content_node = node->children;
	int end = strtol((char *) content_node->content, NULL, 10);
	if (end < 0)
		return;

	cycle *c = alloc(sizeof(cycle));
	c->next = ci->cycle;
	c->position = ci->cycles++;
	c->speed = speed;
	c->start = start;
	c->end = end;
	c->next_step = 1;
	ci->cycle = c;
}

static int compar_cycle(const void *p1, const void *p2) {
	const cycle *c1 = *((const cycle **) p1);
	const cycle *c2 = *((const cycle **) p2);
	float f1 = c1->next_step / (float) c1->speed;
	float f2 = c2->next_step / (float) c2->speed;
	return f1 < f2 ? -1
		: f1 > f2 ? 1
		: c1->position < c2->position ? -1
		: c1->position > c2->position ? 1
		: 0;
}

static key_handler *key_handler_create(process *father) {
	key_handler *k = alloc(key_handler_sizeof());
	process *p = KEY_HANDLER_CAST(k, process);
	process_init(p, father);
	p->run = key_handler_run;
	p->run(p);
	return k;
}

static void key_handler_run(process *p) {
	if (p->state)
		goto *p->state;

	for (;;) {
		if ((key(_ALT) && key_down(_F4)) || key_down(_ESC)
			|| quit_requested) {
			let_me_alone(p);
			break;
		}
		if (key(_ALT) && key_down(_ENTER)) {
			full_screen = !full_screen;
			set_mode(get_screen()->w, get_screen()->h);
		}

		FRAME(p);
	}
	send_signal(p, S_KILL);
}

static size_t key_handler_sizeof(void) {
	return process_sizeof();
}

static char *load_xml(xmlDocPtr document, cycle_init *ci) {
	xmlNodePtr node = document->children;
	if (!node) {
		warnx("empty document");
		return NULL;
	}
	if (!xmlStrEqual(node->name, (const xmlChar *) "palanim")) {
		warnx("Invalid animation");
		return NULL;
	}

	node = node->children;
	if (xmlNodeIsText(node))
		node = node->next;
	if (!xmlStrEqual(node->name, (const xmlChar *) "image")) {
		warnx("Invalid animation, expected image");
		return NULL;
	}
	xmlNodePtr image_node = node->children;
	char *image_file = strdup((char *) image_node->content);

	for (node = node->next; node; node = node->next) {
		if (xmlNodeIsText(node))
			continue;
		if (!xmlStrEqual(node->name, (const xmlChar *) "cycle"))
			continue;
		add_cycle(node->children, ci);
	}
	return image_file;
}

static palanim *palanim_create(process *father, const char *path) {
	xmlDocPtr document = xmlParseFile(path);
	if (!document) {
		warnx("Document not parsed successfully.");
		return NULL;
	}

	cycle_init ci = { 0, NULL };
	char *image_file = load_xml(document, &ci);
	xmlFreeDoc(document);
	if (!image_file)
		return NULL;

	graphic *g;
	if (image_file[0] == '/')
		g = graphic_from_file(image_file);
	else {
		char *absolute_path;
		char *slash = strrchr(path, '/');
		if (slash) {
			size_t dir_length = slash - path + 1;
			size_t image_length = strlen(image_file);
			char *image_path = alloc(dir_length + image_length + 1);
			char *end = mempcpy(image_path, path, dir_length);
			end = mempcpy(end, image_file, image_length);
			*end = '\0';
			absolute_path = canonicalize_file_name(image_path);
			free(image_path);
		}
		else
			absolute_path = canonicalize_file_name(image_file);
		if (!absolute_path) {
			warn("%s", image_file);
			fatal_error();
		}
		g = graphic_from_file(absolute_path);
		free(absolute_path);
	}
	free(image_file);
	const SDL_Surface *surface = get_surface(g);

	palanim *pa_p = alloc(palanim_sizeof());
	process *p = PALANIM_CAST(pa_p, process);
	process_init(p, father);
	p->x = surface->w / 2;
	p->y = surface->h / 2;
	p->graph = g;
	p->on_kill = palanim_on_kill;
	p->run = palanim_run;
	palanim_private *pa = palanim_get_private(pa_p);
	pa->do_cycle = 1;
	pa->cycle = ci.cycle;
	pa->cycle_order = alloc(ci.cycles * sizeof(cycle *));

	SDL_Palette *palette = get_surface(g)->format->palette;
	int ncolors = palette->ncolors;
	pa->ncolors = ncolors;
	SDL_Color *colors = alloc(2 * ncolors * sizeof(SDL_Color));
	pa->colors = colors;
	pa->original_colors = colors + ncolors;
	memcpy(pa->colors, palette->colors, ncolors * sizeof(SDL_Color));
	memcpy(pa->original_colors, palette->colors,
		ncolors * sizeof(SDL_Color));

	pa->ticks = 0;
	p->run(p);
	return pa_p;
}

static palanim_private *palanim_get_private(palanim *pa) {
	void *p = pa;
	return p + private_offset;
}

static void palanim_on_kill(process *p) {
	graphic_destroy(p->graph);
	palanim_private *pa = palanim_get_private((palanim *) p);
	cycle *c = pa->cycle;
	while (c) {
		cycle *next = c->next;
		free(c);
		c = next;
	}
	free(pa->cycle_order);
	free(pa->colors);
}

static void palanim_run(process *p) {
	palanim_private *pa = palanim_get_private((palanim *) p);
	if (key_down(_TAB)) {
		pa->do_cycle = !pa->do_cycle;
		if (!pa->do_cycle) {
			pa->ticks = 0;
			cycle *c;
			for (c = pa->cycle; c; c = c->next)
				c->next_step = 1;
			memcpy(pa->colors, pa->original_colors,
				pa->ncolors * sizeof(SDL_Color));
			convert_palette(p->graph, pa->colors);
		}
	}
	if (pa->do_cycle) {
		pa->ticks++;
		size_t length = 0;
		unsigned fps_7 = 7 * FPS;
		if (pa->ticks >= fps_7) {
			pa->ticks = 0;
			cycle *c;
			for (c = pa->cycle; c; c = c->next) {
				c->next_step = 0;
				pa->cycle_order[length++] = c;
			}
		}
		else {
			unsigned ticks_2 = 2 * pa->ticks;
			cycle *c;
			for (c = pa->cycle; c; c = c->next) {
				if (ticks_2 >= c->next_step * fps_7 / c->speed)
					pa->cycle_order[length++] = c;
			}
		}

		if (length) {
			SDL_Color *colors = pa->colors;
			qsort(pa->cycle_order, length, sizeof(cycle *),
				compar_cycle);
			size_t s;
			for (s = 0; s < length; s++) {
				cycle *c = pa->cycle_order[s];
				int start = c->start;
				int end = c->end;
				SDL_Color temp_color = colors[end];
				if (start < end) {
					int i;
					for (i = end; i > start; i--)
						colors[i] = colors[i - 1];
				}
				else {
					int i;
					for (i = end; i < start; i++)
						colors[i] = colors[i + 1];
				}
				colors[start] = temp_color;
				c->next_step++;
			}
			convert_palette(p->graph, colors);
		}
	}
	frame_done(p);
}

static size_t palanim_sizeof(void) {
	static size_t size;
	if (!size)
		size = private_offset + sizeof(palanim_private);
	return size;
}

static void process_args(int argc, char **argv) {
	if (argc < 2)
		errx(EXIT_FAILURE, "missing file operand");
	palanim_path = argv[1];
}

static void viewer_run(process *p) {
	if (p->state)
		goto *p->state;

	set_title(palanim_path);
	set_fps(FPS, 0);
	set_mode(320, 200);

	palanim *pa = palanim_create(p, palanim_path);
	if (!pa)
		fatal_error();
	process *pa_p = PALANIM_CAST(pa, process);
	const SDL_Surface *surface = get_surface(pa_p->graph);
	if (surface->w != 320 || surface->h != 200)
		set_mode(surface->w, surface->h);

	key_handler_create(p);

	for (;;) {
		FRAME(p);
	}
	send_signal(p, S_KILL);
}

static size_t viewer_sizeof(void) {
	return process_sizeof();
}
